from attributedict.collections import AttributeDict
from pb_common.client import exceptions
from pb_common.utils import support

AGREE_SECTIONS = ('masterAccount', 'cardContracts', 'currencyContracts', 'goalContracts',
                  'walletContracts', 'loanContracts', 'depositContracts')


class Agreement(AttributeDict):
    """
    Договор клиента
    """
    def __init__(self, context, agreement, integration_id):
        self._context = context
        self.integration_id = integration_id
        self._contract_info = None
        self.last_sort_order = -1
        if 'cards' in agreement and agreement['cards']:
            for index, card in enumerate(agreement['cards']):
                agreement['cards'][index] = Card(context, card, integration_id)
            agreement['cards'] = sorted(agreement['cards'], key=lambda c: c.sortOrder)
        super().__init__(agreement)

    @property
    def display_name(self):
        """
        Название договора на дашборде
        """
        if self.type == 'card' and self.cards:
            return self.cards[0].display_name
        elif self.get('prefixDisplayName', None):
            return f'{self.prefixDisplayName} «{self.displayName}»'
        return self.displayName

    @property
    def account_id(self):
        """Номер договора"""
        return self.id

    @property
    def main_amount(self):
        """Главная сумма на дашборде"""
        if self.availableLimitAmount.amount:
            return self.availableLimitAmount.amount / 100
        elif self.debtAmount.amount:
            return self.debtAmount.amount / 100
        return 0
        # TODO: дописать метод!!!

    def show(self, hidden: bool = False) -> bool:
        """
        Отобразить договор на дашборде
        :param hidden: True - скрыть, False - отобразить
        :return: True - успех, False - ошибка
        """
        r = self._context.contract_service.update_contract(
            path=dict(account_id=self.account_id),
            headers={'Authorization': self.integration_id, 'Content-Type': 'application/json'},
            data='{"hidden": %s}' % str(hidden).lower())
        return r.status_code == 200

    @property
    def uuid(self):
        """
        Сформировать UUID по номеру договора и карты (если есть)
        """
        if self.productId in ('VTB_CARD', 'VTB_PPAY') and getattr(self, 'cards', None):
            account_uuid = f'{self.account_id}{self.cards[0].id}'
        else:
            account_uuid = str(self.account_id)
        return account_uuid

    @support.retry_request(delay=5.0, exceptions=(exceptions.ContractInfoError,))
    def advanced(self, online=False):
        """
        Получить информацию из ContractService по договору
        :param online: выполнить запрос в ContractService если ранее данные уже были получены
        :return: ContractInfo
        """
        if not self._contract_info or online:
            r = self._context.contract_service.get_contract(
                params=dict(agreeId=self.id, online=str(online).lower()),
                headers={'Authorization': f'IntegrationId {self.integration_id}', 'Content-Type': 'application/json'}
            )
            if r.status_code != 200:
                raise exceptions.ContractInfoError(f'Не удалось получить данные из Contract Service: {r.text}')
            self.__dict__['_contract_info'] = ContractInfo(r.json())
        return self._contract_info

    @property
    def is_first(self):
        """
        Признак, является ли договор первым в списке отобржения на Дашборде
        :return: True - первый
        """
        return self.sortOrder == 1

    @property
    def is_last(self):
        """
        Признак, является ли договор последним в списке отобржения на Дашборде
        :return: True - последний
        """
        return self.sortOrder == self.last_sort_order


class ContractInfo(AttributeDict):
    """Информация по договору из Contract Service"""

    @property
    def reduce_pay(self):
        """
        Возможно ли подключить услугу - Уменьшение платежа?
        :return: True - возможно, False - невозможно или уже активна
        """
        services = getattr(self, 'services', [])
        for service in services:
            if service['serviceType'] == 'ReducePay':
                return True if service['status'] == 'В' else False
        return False


class Card(AttributeDict):
    """
    Карта к договору клиента
    """
    def __init__(self, context, card, integration_id):
        self._context = context
        self.integration_id = integration_id
        self._card_no = ''
        super().__init__(card)

    @property
    def card_no(self):
        """
        Номер карты - не маскированный
        """
        if self._card_no != '':
            return self._card_no

        v_cards = self._context.vabank_provider.execute(f'SELECT * FROM CRP_CARDS WHERE accountid={self.contractId}')
        for v_card in v_cards:
            if getattr(self, 'cardNo', None):
                self._card_no = None
            if v_card['CARDNO'][-4:] == self.number[-4:]:
                self._card_no = v_card['CARDNO']
        return self._card_no

    @property
    def quartile(self):
        """
        Третья квартиль карты
        """
        if self.card_no:
            return self.card_no[8:12]

    @property
    def display_name(self):
        """
        Название карты на дашборде
        """
        if getattr(self, 'prefixDisplayName', None):
            return f'{self.prefixDisplayName} «{self.displayName}»'
        return self.displayName

    @property
    def uuid(self):
        """
        Сформировать UUID по номеру договора и карты
        """
        return f'{self.contractId}{self.id}'

    @property
    def main_amount(self):
        """Главная сумма на дашборде"""
        if self.availableLimitAmount.amount:
            return self.availableLimitAmount.amount / 100
        elif self.debtAmount.amount:
            return self.debtAmount.amount / 100
        return 0
        # TODO: дописать метод!!!

    @property
    def type(self):
        """
        Тип продукта - Карта
        :return:
        """
        return 'card'

    def show(self, hidden: bool = False) -> bool:
        """
        Отобразить карту на дашборде
        :param hidden: True - скрыть, False - отобразить
        :return: True - успех, False - ошибка
        """
        r = self._context.contract_service.update_contract(
            path=dict(account_id=f'{self.contractId}/card/{self.id}'),
            headers={'Authorization': self.integration_id, 'Content-Type': 'application/json'},
            data='{"hidden": %s}' % str(hidden).lower())
        return r.status_code == 200


class Agreements(list):
    """
    Список договоров клиента
    """
    def __init__(self, context, dashboard):
        self._context = context
        self._dashboard = dashboard
        self._agreements = []

        for group_agreements in AGREE_SECTIONS:
            agreements = self._dashboard.get(group_agreements)
            if not agreements:
                continue
            elif isinstance(agreements, list):
                agrees = [Agreement(context, agr, self.integration_id) for agr in agreements]
            elif isinstance(agreements, dict):
                agrees = [Agreement(context, agreements, self.integration_id)]
            else:
                raise TypeError('Неожиданный тип договора')

            self._agreements.extend(agrees)
            self._dashboard[group_agreements] = agrees

        last_sort_order = max([a.sortOrder for a in self._agreements])
        for a in self._agreements:
            a.last_sort_order = last_sort_order

        super().__init__(self._agreements)

    @property
    def curacc(self):
        """Сбер. счет"""
        current = self._dashboard.get('masterAccount', [])
        if current:
            return current[0]
        return {}

    @property
    def cards(self):
        """Все договора - Карты"""
        cards = self._dashboard.get('cardContracts', [])
        return sorted(cards, key=lambda x: x.sortOrder)

    @property
    def credit_cards(self):
        """Все договора - Кредитные карты"""
        return list(filter(lambda c: c.debit is False, self.cards))

    @property
    def debit_cards(self):
        """Все договора - Дебетовые карты"""
        return list(filter(lambda c: c.debit is True, self.cards))

    @property
    def credits(self):
        """Все договора - Кредиты"""
        credits = self._dashboard.get('loanContracts', [])
        return sorted(credits, key=lambda x: x.sortOrder)

    @property
    def deposits(self):
        """Все договора - Вклады"""
        deposits = self._dashboard.get('depositContracts', [])
        return sorted(deposits, key=lambda x: x.sortOrder)

    @property
    def wallets(self):
        """Все договора - Кошельки"""
        wallets = self._dashboard.get('walletContracts', [])
        return sorted(wallets, key=lambda x: x.sortOrder)

    @property
    def goals(self):
        """Все договора - Цели"""
        goals = self._dashboard.get('goalContracts', [])
        return sorted(goals, key=lambda x: x.sortOrder)

    @property
    def quartile_card(self):
        """
        Получить любую квартиль карты для регистрации
        :return: квартиль карты
        """
        agrees = [self.curacc] if self.curacc else []
        agrees += self.cards
        for agree in agrees:
            if 'cards' in agree and agree['cards']:
                for card in agree.cards:
                    if card.state == 'active' and card.isPrimary:
                        return card.quartile
        raise AttributeError('Не найдено действующих карт к Сбер. счету и карточным продуктам')

    @property
    def integration_id(self):
        """
        Идентификатор клиента из ответа дашборд сервис
        :return: integration_id
        """
        return self._dashboard.session.user.id

    @property
    def account(self):
        """
        Информация о клиенте
        :return:
        """
        return self._dashboard.session.account

    def __getattr__(self, item):
        """
        Получить атрибуты из ответа DashboardService из переменной self._dashboard
        :param item: название атрибута
        :return: значение
        """
        try:
            return self._dashboard[item]
        except KeyError:
            raise KeyError(f'Не найден атрибут в ответе Dashboard сервиса: {item}')
