import os

import requests
import yaml
from config2.config import Config, ConfigError, DEFAULT_CONFIG_SERIALIZERS, DEFAULT_ENV_KEYS, DEFAULT_ENV


class GetConfigException(Exception):
    pass


class GetCertsException(Exception):
    pass


class ConfigRu(Config):
    @staticmethod
    def load_file(config_file):
        config_file = config_file or {}  # AttributeDict({})

        try:
            if hasattr(config_file, 'path') and config_file['path']:
                with open(config_file.path, encoding='utf8') as file:
                    try:
                        config_file.raw = file.read()

                    except IOError as error:
                        raise ConfigError(
                            'Tried but failed to `read` detected config file `{path}`.'.format(**config_file))

                    try:
                        serializer = DEFAULT_CONFIG_SERIALIZERS[config_file.format]

                        config_file.data = serializer.unpack(config_file.raw)

                    except Exception as error:
                        raise ConfigError('Tried but failed to `unpack` (de-serialize) detected '
                                          'config file `{path}` using serializer `{format}`.'.format(**config_file))

        except AttributeError as error:
            raise ConfigError(
                'Tried but failed to `load` detected config file.'.format(**config_file) + ' Reason: {}'.format(error))

        return config_file

    @staticmethod
    def create(*args, **kwargs):
        env = dict(enumerate(args)).get(0, None)
        env = env or kwargs.get('env', None)
        env = env or Config.detect_env(DEFAULT_ENV_KEYS, DEFAULT_ENV)

        try:
            del args[0]
        except:
            pass

        try:
            del kwargs['env']
        except:
            pass

        cfg = ConfigRu(env, *args, **kwargs)

        return cfg


class ConfigCreator(object):
    DEFAULT = 'default'

    def __init__(self, config):
        self._config = config
        self._ext = ('.yaml', '.yml')
        self._get_config = config.get_config
        self._current_env = os.environ.get('ENV', None)
        self._merge_strategy = os.environ.get('MC_STRATEGY', 'LOCAL').upper()

    @property
    def get_configs_url(self):
        """
        Базовый URL для запроса конфига в сервис-конфигов
        :return: url
        """
        return self._get_config.url + self._get_config.get_env

    @property
    def get_certs_url(self):
        """
        Базовый URL для запроса сертификата в сервис-конфигов
        :return: url
        """
        return self._get_config.url + self._get_config.get_cert

    def get_environment_url(self, env=None):
        """
        Получить URL для доступа к конфигу окружения
        :param env: окружение
        :return: url
        """
        if env:
            return self.get_configs_url + '/' + env
        elif self._current_env:
            return self.get_configs_url + '/' + self._current_env
        return self.get_configs_url

    def get_config(self, env=None):
        """
        Получить конфиг из сервиса конфигов
        :return: данные конфигов
        """
        url = self.get_environment_url(env)
        try:
            response = requests.get(url=url, timeout=30)
        except (requests.Timeout, requests.ReadTimeout, requests.ConnectTimeout, requests.ConnectionError):
            raise requests.ConnectionError(f'Сервис конфигов недоступен: {url}')
        if response.status_code == 200:
            cfg = response.json()
            if env or self._current_env:
                return {env or self._current_env: cfg}
            return cfg
        raise GetConfigException(response.text)

    def get_configs(self):
        """
        Получить конфиг окружения и default-конфиг
        :return: конфиг
        """
        config = self.get_config()
        if self._current_env:
            default_config = self.get_config(env=self.DEFAULT)
            config.update(default_config)
        return config

    def _has_local_configs(self, name):
        """
        Проверить наличие файла локального конфига
        :param name: имя конфига
        :return: путь до файла локального конфига, если он существует
        """
        config_path = self._config.__config_path__
        for ext in self._ext:
            if os.path.isfile(os.path.join(config_path, name + ext)):
                return os.path.join(config_path, name + ext)

    def _merge_local_config(self, name, data):
        """
        Смерджить данные существующего локального конфига с данными от конфиг-сервиса
        :param name: имя конфига
        :param data: данные конфига
        """
        local_config_file = self._has_local_configs(name)
        with open(local_config_file, encoding='utf8') as stream:
            local_data = yaml.safe_load(stream)
        if self._merge_strategy == 'LOCAL':
            merged_data = Config.merge(data, local_data)
        else:
            merged_data = Config.merge(local_data, data)
        with open(local_config_file, 'w', encoding='utf8') as stream:
            yaml.safe_dump(merged_data, stream, indent=4, allow_unicode=True)

    def _create_local_config(self, name, data):
        """
        Создать новый файл локального конфига
        :param name: имя конфига
        :param data: данные конфига
        """
        config_path = self._config.__config_path__
        local_config_file = os.path.join(config_path, name + self._ext[0])
        with open(local_config_file, 'w', encoding='utf8') as stream:
            yaml.safe_dump(data, stream, allow_unicode=True)

    def _create_local_cert(self, name, data):
        """
        Создать новый файл локального сертификата
        :param name: имя сертификата
        :param data: данные сертификата
        """
        config_path = self._config.__config_path__
        local_cert_file = os.path.join(config_path, name)
        with open(local_cert_file, 'w', encoding='utf8') as stream:
            stream.write(data)

    def get_cert(self, env):
        """
        Получение сертификата для контура
        :param env: контур
        """
        url = self.get_certs_url + '/' + env
        try:
            response = requests.get(url=url, timeout=30)
        except (requests.Timeout, requests.ReadTimeout, requests.ConnectTimeout, requests.ConnectionError):
            raise requests.ConnectionError(f'Сервис конфигов недоступен: {url}')
        if response.status_code == 200:
            return response.json()
        raise GetCertsException('Ошибка при запросе сертификатов {}: {}'.format(env, response.text))

    def get_certs(self):
        """
        Получить сертификаты
        """
        certs = self.get_cert(self.DEFAULT)
        if self._current_env:
            certs_env = self.get_cert(self._current_env)
            certs.update(certs_env)
        for name, value in certs.items():
            self._create_local_cert(name, value)

    def reload(self):
        """
        Перегрузить конфиг-файлы для текущего окружения ENV
        :return: config
        """
        configs = self.get_configs()
        for config_name, config_data in configs.items():
            if self._has_local_configs(config_name):
                self._merge_local_config(config_name, config_data)
            else:
                self._create_local_config(config_name, config_data)
        self.get_certs()
        os.chdir(self._config.__root_path__)
        self._config.reload()
        return self._config
