from functools import partial

import redis
from redis.sentinel import Sentinel

from pb_common.db import AbstractDbProvider
from pb_common.utils.logger import get_logger

logger = get_logger(__name__)


class UnknownMapping(Exception):
    pass


class RedisProvider(AbstractDbProvider):
    def __init__(self, hosts, master_name, password):
        self.hosts = hosts
        self.master = master_name
        self.password = password
        self.sentinel = Sentinel(self.hosts)
        self.client = None

    def connect(self):
        logger.debug('Connect to Redis...')
        host, port = self.sentinel.discover_master(self.master)
        self.client = redis.StrictRedis(host=host, port=port, password=self.password)
        logger.debug('Connected: {}'.format(self.client))
        return self

    @property
    def get_client(self):
        if not self.client:
            self.connect()
        return self.client

    def __getattr__(self, item):
        logger.debug('Execute method: {}'.format(item))
        redis_method = getattr(self.client, item)
        return partial(self.execute, redis_method)

    def execute(self, method, *args, **kwargs):
        logger.debug('Params to execute: {} {}'.format(args, kwargs))
        r = method(*args, **kwargs)
        logger.debug('Response: {}'.format(r))
        return r

    def custom_get(self, key):
        mapping = dict(
            string=self.client.get,
            hash=self.client.hgetall,
        )
        logger.debug('Params to execute: {}'.format(key))
        exist_key = self.client.exists(key)
        if exist_key > 0:
            key_type = self.client.type(key)
            try:
                return mapping[key_type.decode('utf8') if isinstance(key_type, bytes) else str(key_type)](key)
            except KeyError:
                raise UnknownMapping('Неподдерживаемый тип ключа: {}. Добавьте в mapping новый метод'.format(key_type))
        return False

    def close(self):
        if self.client:
            self.client.close()
        logger.debug('Closed Redis connection.')
