import ssl
from copy import deepcopy
from functools import partial
from http.client import HTTPSConnection
from urllib.parse import quote_plus, urlparse

from pb_common.services import AbstractRequestProvider
from pb_common.utils.logger import get_logger

__author__ = "Maksim Senchuk"

logger = get_logger(__name__)


class MethodUndefined(Exception):
    pass


class HttpsProvider(AbstractRequestProvider):
    def __init__(self, service: str, protocol: str, host: str, port: int, methods: dict, path=None, headers=None):
        self.service_name = service
        self._protocol = protocol
        self._host = host
        self._port = port
        self._methods = methods
        self._path = path
        self._headers = dict(headers) if headers else {}

    def base_url(self):
        url = "{protocol}://{host}:{port}".format(protocol=self._protocol, host=self._host, port=self._port)
        if self._path:
            url += self._path if self._path.startswith('/') else '/' + self._path
        return url

    def __getattr__(self, item):
        if dict(self._methods).get(item):
            http_method = self._methods[item][0]
            url_path = self._methods[item][1]
            url = self.base_url() + (url_path if url_path.startswith('/') else '/{}'.format(url_path))
            return partial(self.send, method=http_method, url=url)
        else:
            raise MethodUndefined("Не найден {} метод {} сервиса в списке разрешенных методов: {}".format(
                item, self.service_name, self._methods))

    def send(self, method, url, path: dict = None, params: dict = None, quote: bool = False,
             protocol=ssl.PROTOCOL_TLSv1_2, timeout=60, **kwargs):
        """
        Отправить запрос
        :param method: метод запроса GET/POST/PUT и т.д
        :param url: урл куда отправлять запрос
        :param path: словарь ключей и значений, где ключи это динамические данные в урл
        :param params: словарь ключ-значение, который добавится в урл в виде ?ключ1=значение1&ключ2=значение2
        :param quote: bool - переводить ли строку поле ? в url-строку
        :param protocol: версия SSL/TLS протокола
        :param timeout: время ожидания ответа
        :param kwargs: остальные параметры запроса
        :return: Response
        """
        if kwargs.get('headers'):
            h = deepcopy(self._headers)
            h.update(kwargs['headers'])
            kwargs['headers'] = h
        else:
            kwargs['headers'] = self._headers
        if not kwargs['headers'].get('User-Agent') or not kwargs['headers'].get('user-agent'):
            kwargs['headers'].update({'User-Agent': 'AutoTest User Agent by: {}'.format(__author__)})
        if path:
            url = url.format(**path)
        if quote and params:
            params = {k: quote_plus(v) for k, v in params.items()}
        if params:
            url += '?' + '&'.join([f'{k}={v}' for k, v in params.items()])
        logger.info("Send to '{}' request: {} {}".format(self.service_name, method, url))

        url_parsed = urlparse(url)
        ctx = ssl.SSLContext(protocol=protocol)
        try:
            conn = HTTPSConnection(url_parsed.hostname, context=ctx)
            conn.connect()

            req_url = url_parsed.path
            if url_parsed.query:
                req_url += f'?{url_parsed.query}'

            conn.request(method, req_url, **kwargs)
            r = conn.getresponse()
            content = r.read()
        finally:
            conn.close()
        setattr(r, 'content', content)
        setattr(r, 'text', str(content, 'utf8'))
        logger.info("Response: {} {}\nHeaders: {}\nBody: {}".format(r.code, r.reason, r.headers, r.text))
        return r
