from datetime import datetime
from functools import wraps

import allure
from allure_commons.utils import func_parameters, represent
from allure_commons._allure import StepContext
from allure_commons import plugin_manager
from allure_commons.types import AttachmentType

from pb_common.utils.logger import get_logger

logger = get_logger(__name__)


def attach(name, body='', attachment_type=AttachmentType.TEXT):
    msg = 'Allure attachment:\n{}:\n{}'.format(name, body)
    if attachment_type != AttachmentType.TEXT:
        msg = 'Allure attachment: {} -> {}'.format(name, attachment_type)
    logger.debug(msg)
    allure.attach(body, name, attachment_type)


def allure_testcase(case_id, name='Ссылка на тест-кейс в TestRail'):
    return allure.testcase(case_id, f'{name}: {case_id}')


def step(title):
    if callable(title):
        return StepContextWrap(title.__name__, {}, True)(title)
    else:
        return StepContextWrap(title, {})


def strict_step(func_strict: bool = True):
    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                if func_strict:
                    raise e

        return wrapper
    return decorator


class StepContextWrap(StepContext):

    def __init__(self, title, params, is_callable=False):
        super().__init__(title, params)
        self.is_callable = is_callable
        self.start_time = datetime.today().strftime("%Y-%m-%d %H:%M:%S.%f")

    def log_step(self, action, title=None):
        if title is None:
            title = self.title
        logger.info('{} step: {}'.format(action, title))

    def _allure_step(self, title=None):
        if title is None:
            title = self.title
        return "[{}] {}".format(self.start_time, title)

    def __enter__(self):
        plugin_manager.hook.start_step(
            uuid=self.uuid,
            title=self._allure_step(),
            params=self.params
        )
        if self.is_callable is False:
            self.log_step('Start')

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            logger.error('STEP FAILED! Reason:\n{}'.format(exc_val))
        if self.is_callable is False:
            self.log_step('End')

        plugin_manager.hook.stop_step(
            uuid=self.uuid,
            title=self._allure_step(),
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb
        )

    def __call__(self, func):
        @wraps(func)
        def impl(*a, **kw):
            __tracebackhide__ = True
            params = func_parameters(func, *a, **kw)
            args = list(map(lambda x: represent(x), a))
            title = self.title.format(*args, **params)
            with StepContextWrap(title, params):
                return func(*a, **kw)

        return impl
